-- console.vhd
-- ------------------------------------------------------
--  Serial Terminal Emulator with File I/O  (compile -93)
-- ------------------------------------------------------
-- (c) ALSE - cannot be used without the prior written consent of ALSE
-- Author : Bert Cuzeau
-- E-mail : info@alse-fr.com
-- Web    : http://www.alse-fr.com
--
-- Reads in characters from the file which name
-- is passed as a generic parameter, and sends
-- them over the RS232 line.
-- When the end of file is reached, the ENDF signal
-- is raised.
-- Note : the Serial Output is name "Rx" (it is to be
-- connected to the Rx input of UARTS).
--
-- Text File syntax :
--   ; <-- comments (rest of line is ignored)
--   #200  <-- delay (pause in transmission expressed in us)
-- Note : end_of_lines chars (Cr & Lf) are not sent.

  use std.textio.all;
LIBRARY ieee;
  USE ieee.std_logic_1164.ALL;
  USE ieee.numeric_std.all;

-- ----------------------------------------------------
    Entity CONSOLE is
-- ----------------------------------------------------
    Generic ( BaudRate : integer range 1200 to 460800;
              filename : string  := "console.txt";
              Parity   : boolean := false;
              Even     : boolean := false
            );
    Port    ( RX  : out std_logic;        -- Serial output
              CTS : in  std_logic := '1'; -- if CTS='0', we stop sending
              ENDF: out boolean);
end CONSOLE;

-- ----------------------------------------------------
    Architecture behavioral of CONSOLE is
-- ----------------------------------------------------

  constant BITPERIOD : time := 1E9 ns / BaudRate;
  constant NDbits    : positive := 8;  -- Nb of DATA bits. Do not modify.

  signal RSData : std_logic_vector (7 downto 0);
  file   F : text;

begin

  process
    variable L  : line;
    variable c  : character;
    variable ok : boolean;
    variable k  : integer := 0;
    variable ParityBit : std_logic;
  begin

    ENDF <= FALSE;
    RX <= '1';          -- Idle

    file_open(F, filename, read_mode);

    wait for BITPeriod * 1.6; -- Startup delay

    while not ENDFILE(F) loop

      readline (F,L);
      -- interpret the data line
      while L'length /= 0 loop
        if L(1)=';' then -- skip comments
          exit;
        elsif L(1)='#' then
          read (L,c);    -- skip the #
          read (L,k,ok);
          if Ok then
            wait for (k * 1 us);
          end if;
          exit;
        else
          -- send this character
          read (L,c);

          -- do not send if outside says busy (RTS=0)
          if CTS='0' then       -- remember : "wait until" is edge sensitive...
            wait until CTS='1';
          end if;

          -- convert character to std_logic_vector :
          RSdata <= std_logic_vector(to_unsigned(character'pos(c),8));

          RX <= '0';     -- Start bit
          ParityBit := '0';
          wait for BITperiod;

          for i in 0 to NDbits-1 loop -- Data bits
            ParityBit := ParityBit xor RSData(i);
            RX <= RSData(i); wait for BITperiod;
          end loop;

          if Parity then  -- Parity bit if necessary
            if not Even then
              ParityBit := not ParityBit;
            end if;
            RX <= ParityBit; wait for BITperiod;
          end if;

          RX <= '1';     -- 1.5 x Stop bits
          wait for BITperiod * 1.5;
        end if;
      end loop;
    end loop;

    file_close (f);

    -- make sure all has been sent and echoed.
    wait for (24 * BITperiod);
    ENDF <= TRUE;
    wait;  -- kill this process

  end process;

end behavioral;
