-- tb_top.vhd
-- ---------------------------------------------
--    VHDL Test Bench for UART Top_Level
-- ---------------------------------------------
-- (c) ALSE - Bert Cuzeau
-- e-mail :  info@alse-fr.com
-- Web    : http://www.alse-fr.com
-- Compile in '93
--
-- Note that we've put the Constants in the entity section.
-- You must match the BaudRate here with the UARTS combination
-- of Baud1/Baud2 (in TOP_UART) and DIPSWitch position (here).
-- Xtal freq and parity settings are passed all the way down
-- the hierarchy, automatically.
--
-- During the simulation, you don't really need the waveform :
-- the activity is logged in the transcript : character echoed
-- by the system are "received" in this testbench and written
-- in the simulation transcript.

  USE std.textio.all;
LIBRARY ieee;
  USE ieee.std_logic_1164.ALL;
  USE ieee.numeric_std.ALL;
  USE ieee.std_logic_textio.ALL;

-- ---------------------------------------------
    Entity TB_TOP is
-- ---------------------------------------------
-- Enter HERE all the parameters :
-- -----------------------------------
  constant Fxtal    : integer  := 33333300; -- 33.3333 MHz NIOS board
--constant Fxtal    : integer  := 40000000; -- 40 MHz : Insight Board
--constant Fxtal    : integer  := 50000000; -- 50 MHz
-- -----------------------------------
  constant BaudRate : integer  := 115200;
  constant Parity   : boolean  := false;
  constant Even     : boolean  := false;
  constant RstPol   : std_logic:= '0';
-- -----------------------------------
end TB_TOP;

-- ---------------------------------------------
    Architecture TEST of TB_TOP is
-- ---------------------------------------------

  constant Period    : time := 1 sec / Fxtal;  -- System Clock
  constant BITPeriod : time := 1 sec / BaudRate;

  signal CLK   : std_logic := '0';
  signal nRST  : std_logic;
  signal DIPSW : std_logic;
  signal RX    : std_logic;
  signal TX    : std_logic;
  signal ENDF  : boolean;

begin

  -- --------------------------
  --  UUT Instanciation
  -- --------------------------
  UUT : entity work.TOP_UART
  Generic map ( Fxtal     => Fxtal,
                Parity    => Parity,
                Even      => Even,
                Reset_Active => RstPol )
     Port Map ( CLK       => CLK,
                Reset_n   => nRST,
                DIPSW(0)  => DIPSW,
                RS232_RX0 => RX,
                RS232_TX0 => TX );

  -- -----------------------------
  --  Clock, Reset & DIP-Switches
  -- -----------------------------

  nRST <= RstPol, not RstPol after Period;

  CLK <= '0' when ENDF else not CLK after (Period / 2);

  -- DIPSW<='0'; -- ->  19.200 bauds
  DIPSW<='1'; -- -> 115.200 bauds

  -- --------------------------
  --  Console
  -- --------------------------

  -- Generic "filename" can be set by the configuration section
  i_CONSOLE : entity work.CONSOLE
   Generic Map ( BaudRate=> BaudRate,
                 Parity  => Parity,
                 Even    => Even )
    Port Map   ( RX      => RX,
                 CTS     => open,
                 ENDF    => ENDF    );

  assert NOT ENDF  -- note : this is a concurrent "assert"
    report "End of Simulation." severity note;

  -- -----------------------------
  --  Serial Receive (behavioral)
  -- -----------------------------

  process
    variable L   : line;
    variable MOT : std_logic_vector (7 downto 0);
    variable ParityBit : std_logic;
  begin

    loop
      ParityBit := '0';

      wait until TX = '0';         -- get falling edge

      wait for (0.5 * BITperiod);  -- Middle of Start bit
      assert TX = '0'
        report "Error during Start Bit ???" severity warning;

      wait for BITperiod;          -- First Data Bit
      for i in 0 to 7 loop         -- Get word
        MOT(i) := TX;
        ParityBit := ParityBit xor TX;
        wait for BITperiod;
      end loop;

      if Parity then
        if not Even then
          ParityBit := not ParityBit;
        end if;
        if ParityBit /= TX then
          report "Error during Parity Bit" severity warning;
        end if;
        wait for BITperiod;
      end if;

      wait for BITperiod;         -- Stop bit
      assert TX = '1'
        report "Error during Stop bit ???" severity warning;


      write  (L,string'("Character received (hex) = "));
      hwrite (L,MOT);             -- trace
      write  (L,string'(" - '" & character'val(to_integer(unsigned(MOT))) & "'" ));
      writeline (output,L);       -- write to simulation transcript
    end loop;
  end process;

end TEST;
