-- APPLIC.vhd
-- ----------------------------------------------------
--   Communication & "Scrambling" Engine
--   This version prints a prompt ! (modified for Quartus II)
-- ----------------------------------------------------
-- (c) ALSE
-- e-mail : info@alse-fr.com
-- Web    : http://www.alse-fr.com
--
-- Notes :
--
--  * Architecture RTL does some scrambling by adding a changeable
--    offset (0 to 9) to each character received before echoing it.
--    Any single-digit number received becomes the offset.
--  Example : "aaa2aa1a" echoes "aaa.cc.b"
--
--  * There is no buffering. If the sender is not ready to receive
--    the echo, then the data may be lost.
--    There is room only for one character received while we
--    were busy waiting...
--
-- This is only interesting as a trivial example on "How to Use the Uart".
-- A real-world application should probably add more buffering, timeouts...

LIBRARY ieee;
  USE ieee.std_logic_1164.ALL;
  USE ieee.numeric_std.ALL;

-- ----------------------------------------------------
    Entity APPLIC is
-- ----------------------------------------------------
      Port (     CLK : In  std_logic;
                 RST : In  std_logic;
                 RTS : In  std_logic;                    -- from RS232 Flow Control
               RxErr : In  std_logic;                    -- not used
               RxRDY : In  std_logic;                    -- from UARTS
                SDin : In  std_logic_vector (7 downto 0);-- from UARTS
              TxBusy : In  std_logic;                    -- fom UARTS
            LD_SDout : Out std_logic;                    -- to UARTS
               SDout : Out std_logic_vector (7 downto 0) -- to UARTS
            );
end APPLIC;


-- ----------------------------------------------------
    Architecture RTL of APPLIC is
-- ----------------------------------------------------

-- commented out due to Quartus II lack of support for Character'pos !
-- function Char2SLV8 (c : character) return std_logic_vector is
-- begin
--    return std_logic_vector(to_unsigned(character'pos(c),8));
-- end function;

  subtype Byte is std_logic_vector (7 downto 0);
  type Hex_string is array (natural range <>) of Byte;

--constant Prompt : string := "Scrambler Demo." & Cr & Lf;

  -- For Quartus II lack of support for Character'pos ! :
  constant Prompt_hex : Hex_string
                      := ( x"53", x"63", x"72", x"61", x"6D", x"62", x"6C", x"65",
                           x"72", x"20", x"44", x"65", x"6D", x"6F", x"2E", x"0D",
                           x"0A" );

  type State_Type is  (Boot1, Boot2, Boot3, Idle, Scramble, Send);
  signal State : State_Type;

  signal RData : std_logic_vector (7 downto 0);
  signal SData : std_logic_vector (7 downto 0);

  signal Code  : unsigned(3 downto 0);
  signal RTS_r : std_logic;             -- resync FLipFlop on RTS
  signal CharCount : natural range Prompt_hex'low to Prompt_hex'high+1;
  signal CharAvail : boolean;

begin

SDout <= SData;

process (CLK,RST)
begin
  if RST='1' then
    State <= Boot1;
    LD_SDout <= '0';
    SData <= (others=>'0');
    RData <= (others=>'0');
    Code <= (others=>'0');
    RTS_r <= '0';
    CharCount <= Prompt_hex'low;
    CharAvail <= false;

  elsif rising_edge(CLK) then

    -- memorize any arriving character
    if RxRDY='1' then  -- memorize the incoming char
      RData <= SDin;
      CharAvail <= true;
    end if;

    RTS_r <= RTS;
    LD_SDout <= '0';

    case State is

      when Boot1 =>
        CharCount <= Prompt_hex'low;
        State <= Boot2;

      when Boot2 =>
        if (TxBusy='0') and (RTS_r='1') then
          -- SData <= Char2SLV8 (Prompt(CharCount));
          SData <= Prompt_hex(CharCount);
          LD_SDout <= '1';
          CharCount <= CharCount+1;
          if CharCount=Prompt_hex'high then
            State <= Idle;
          else
            State <= Boot3;
          end if;
        end if;

      when Boot3 => -- we need one clock cycle before testing Busy
        State <= Boot2;

      when Idle =>
        if CharAvail then
          CharAvail <= false;
          State <= Scramble;
        end if;

      when Scramble =>
        if (TxBusy='0') and (RTS_r='1') then
          State <= Send;
          case to_integer(unsigned(RData)) is
            when 10|13|32 =>   -- do not translate Cr Lf Space !!!
              SData <= RData;
            when 48 to 57 =>   -- change the offset & returns a dot
              Code <= unsigned(RData(3 downto 0));  -- x"30"..x"3F"
              SData <= x"2E"; -- Char2SLV8('.');
            when 63 => -- '?' -> Prompt
              State <= Boot1; -- override Send
            when others =>     -- add the current offset
              SData <= std_logic_vector(unsigned(RData) + Code);
          end case;
        end if;

      when Send =>
        LD_SDout <= '1';
        State <= Idle;

      -- if encoding is safe + binary and the synthesis tool accepts it...
      when others =>
        State <= Idle;

    end case;

  end if;
end process;


end RTL;

